/*
 *	SunTroff -  A program to display the output of Device Independent
 *		Troff as a window on a Sun Workstation.
 *
 *	Troff_draw.c - Code to do the draw the simple stuff in the page
 *		bit map.
 *
 *	Authors - The original version of this program was written by
 *			Richard L. Hyde (Purdue)
 *			David Slattengren (Berkeley)
 *		It was nearly all rewritten, cleaned up and a more elegant
 *		user interface installed by
 *			Malcolm Slaney (Schlumberger Palo Alto Research)
 *		Combine characters into words
 *			David Cahlander (Cray Research, Inc.)
 *
 *	Legalese -  This command was developed as an independent project
 *		to satisfy a need of the author.  This program may contain
 *		bugs and the user is cautioned to independently verify that
 *		the program is suitable for the user's intended purpose.
 *		The program is made available on an ``as is'' basis with
 *		all faults and without any implied or expressed warranties
 *		or support from either the author, Malcolm Slaney, or
 *		Schlumberger Palo Alto Research Laboratory.
 *
 *		I am putting this program in the public domain.  You are
 *		free to use it as you wish.  In return I ask two things.
 *		First, that you do not remove the names of the authors
 *		from this work.  Secondly, if you make changes or
 *		improvements to this program that you pass these back to
 *		the author so that everybody can benefit from the
 *		improvements.
 *
 *					Malcolm Slaney  (December 1986)
 *					Schlumberger Palo Alto Research
 *					3340 Hillview Avenue
 *					Palo Alto, CA 94304
 *					(415) 496-4669
 *					spar!malcolm@decwrl.dec.com
 *					malcolm@ecn.purdue.edu
 *					malcolm@spar.slb.com (Someday)
 */

#include	"suntroff.h"

#include <X11/Intrinsic.h>
#include <math.h>
extern Window DrawingCanvas;
extern Widget CanvasWidget;
extern Display *dpy;
extern GC gc;
extern GC fillgc;
extern GC cleargc;
extern Pixmap PagePixRect;
extern int ViewTop, ViewLeft, ViewHeight, ViewWidth;
extern int CurrentTop, CurrentLeft;
extern	int	DisplayOutput;		/* Display the output for user? */

#define	XScale(x)	(((x) * sunres) / DeviceResolution - CurrentLeft)
#define	YScale(y)	(((y) * sunres) / DeviceResolution - CurrentTop)

Scale(x)
int	x;
{
	return(((x) * sunres) / DeviceResolution);
}

ClearPage()
{
	if (debug(DRAWDEBUG))
		printf("clearing page\n");
	XFillRectangle(dpy, PagePixRect, cleargc, 0, 0, 
	 (unsigned int) PAGE_PIXEL_WIDTH, (unsigned int) PAGE_PIXEL_HEIGHT);
}

RefreshPage(){
	if (debug(DRAWDEBUG))
		printf("refreshing page\n");
	RepaintCanvas(CanvasWidget, (caddr_t) NULL, (XEvent *) NULL);
}

DrawString(x,y,s)
int	x, y;
char	*s;
{
	if (!CurrentFont)
		fatal("Don't have a Current Font.");

	if (!CurrentFont->Bits)
		fatal("Don't have the bits for the Current Font.");

	XDrawString(dpy, PagePixRect, gc, XScale(x), YScale(y), s, strlen(s));
}

DrawCharacter(x,y,c)
int	x, y, c;
{
	static char s[2] = " ";
	int dx = 0;
	
	if (!CurrentFont){
		fatal("Don't have a Current Font.");
	}

	if (!CurrentFont->Bits){
		fatal("Don't have the bits for the Current Font.");
	}

	if (DisplayOutput) {
		s[0] = c;
		XDrawString(dpy, PagePixRect, gc, 
		 XScale(x) + dx, 
		 YScale(y), s, 1);
	}
}


DrawLine(x,y)
{
	if (DisplayOutput) {
		if (debug(DRAWDEBUG))
			printf("Drawline to %d, %d\n", XScale(HorizontalPosition + x), YScale(VerticalPosition + y));
		XDrawLine(dpy, PagePixRect, gc, 
			  XScale(HorizontalPosition),
			  YScale(VerticalPosition),
			  XScale(HorizontalPosition+x),
			  YScale(VerticalPosition+y));
	}
	HorizontalPosition += x;
	VerticalPosition += y;
}

box(x1, y1, x2, y2)
float	x1, y1, x2, y2;
{
	int i1, j1, i2, j2;

	i1 = round(ditsiz * x1);
	j1 = round(ditsiz * y1);
	i2 = round(ditsiz * x2);
	j2 = round(ditsiz * y2);
	XFillRectangle(dpy, PagePixRect, gc, 
	    XScale(HorizontalPosition) + i1,
	    YScale(VerticalPosition) + j1,
	    (unsigned int) (i2 - i1), (unsigned int) (j2 - j1));
}

line(x1, y1, x2, y2)
float	x1, y1, x2, y2;
{
	int i1, j1, i2, j2;

	i1 = round(ditsiz * x1);
	j1 = round(ditsiz * y1);
	i2 = round(ditsiz * x2);
	j2 = round(ditsiz * y2);
	XDrawLine(dpy, PagePixRect, gc, 
	    XScale(HorizontalPosition) + i1,
	    YScale(VerticalPosition) + j1,
	    XScale(HorizontalPosition) + i2,
	    YScale(VerticalPosition) + j2);
}

round(x)
float	x;
{
	if (x < 0)
		return(x - .5);
	else
		return(x + .5);
}

draw_fraction(num, den)
char num, den;
{
	char s[2];
	int savedsize;

	savedsize = cursize;
	SetFontSize(round(0.55 * cursize));
	LoadFontBits();
	s[0] = num;
	XDrawString(dpy, PagePixRect, gc,
	 XScale(HorizontalPosition), 
	 YScale(VerticalPosition) - round(0.35 * savedsize), s, 1);
	s[0] = den;
	XDrawString(dpy, PagePixRect, gc,
	 XScale(HorizontalPosition) + round(0.6 * savedsize),
	 YScale(VerticalPosition), s, 1);
	line(.2, 0., .6, -.9);
	SetFontSize(savedsize);
	LoadFontBits();
}


point(x,y)
int	x, y;
{
	if (DisplayOutput)
		XDrawPoint(dpy, PagePixRect, gc, x, y);
}

/*
 *  draw a circle
 *	x1 - x position in character box (0. - 1.)
 *	y1 - y position in character box (0. - 1.)
 *	r  - radius of arc (0. - 1.)
 *	fill - TRUE if filled circle
 */

circle(x1, y1, r, fill)
float	x1, y1, r;
int	fill;
{
	arc(x1, y1, r, 0, 360, fill);
}

/*
 *  draw an arc
 *	x1 - x position in character box (0. - 1.)
 *	y1 - y position in character box (0. - 1.)
 *	r  - radius of arc (0. - 1.)
 *	a1 - start angle of arc (degrees)
 *	a2 - length of arc (degrees)
 *	fill - TRUE if filled arc
 */

arc(x1, y1, r, a1, a2, fill)
float	x1, y1, r;
int	a1, a2;
int	fill;
{
	int i1, j1, r1;

	i1 = round(ditsiz * x1);
	j1 = round(ditsiz * y1);
	r1 = round(ditsiz * r);
	if (fill)
		XFillArc(dpy, PagePixRect, fillgc,
		    XScale(HorizontalPosition) + i1 - r1,
		    YScale(VerticalPosition) + j1 - r1,
		    (unsigned int) (2 * r1), (unsigned int) (2 * r1),
		    a1 * 64, a2 * 64);
	else
		XDrawArc(dpy, PagePixRect, gc,
		    XScale(HorizontalPosition) + i1 - r1,
		    YScale(VerticalPosition) + j1 - r1,
		    (unsigned int) (2 * r1), (unsigned int) (2 * r1),
		    a1 * 64, a2 * 64);
}
